/* $Id: x86.c,v 1.1 2004/01/11 22:07:12 mikpe Exp $
 * x86-specific code.
 *
 * Copyright (C) 1999-2004  Mikael Pettersson
 */
#include <stdio.h>
#include <stdlib.h>
#include "libperfctr.h"
#include "arch.h"

static int am_i_k7()
{
    switch (info.cpu_type) {
    case PERFCTR_X86_AMD_K7:
    case PERFCTR_X86_AMD_K8:
    case PERFCTR_X86_AMD_K8C:
        return 1;
    default:
        return 0;
    }
}

extern int decode_level;

struct decode_event_stuff
{
    FILE *fp;
    unsigned evntsel;
};

#ifndef __x86_64__
#define ESCR(obj, n) ((obj)->p4.escr[n])
#else
static unsigned __dummy_escr;
#define ESCR(obj, n) __dummy_escr
#endif

static int decode_event(const struct perfctr_event_set *set,
			const struct perfctr_event *evt,
			void *thunk)
{
    const struct decode_event_stuff *stuff =
	(const struct decode_event_stuff *) thunk;

    if (evt->evntsel == stuff->evntsel) {
	switch (decode_level) {
	case 0:
	    fprintf(stuff->fp, "%02x", evt->evntsel);
	    break;
	case 1:
	    fputs(evt->name, stuff->fp);
	    break;
	case 2:
	    fputs(evt->description ? evt->description : evt->name, stuff->fp);
	    break;
	}
	return -1;
    }

    return 0;
}

static int generic_decode_event(FILE *fp,
				unsigned evntsel, unsigned escr)
{
    fprintf(fp, "event 0x%08X", evntsel);
    if (!am_i_k7() && escr)
	fprintf(fp, "/0x%08X", escr);
    return 0;
}

static int symbolic_decode_event(FILE *fp, unsigned evntsel, unsigned escr)
{
    const struct perfctr_event_set *set;
    struct decode_event_stuff stuff = { fp, evntsel & 0xff };

    if ((set = perfctr_cpu_event_set(info.cpu_type)) == NULL) {
        perror("perfctr_cpu_event_set");
	return -1;
    }

    if (perfctr_walk_event_set(set, decode_event, &stuff) == 0) {
	fprintf(fp, "event 0x%02x", stuff.evntsel);
    }

    if (evntsel & 0x0000ff00) {
	fprintf(fp, "unit=%02x", (evntsel & 0x0000ff00) >> 8);
    }
    if (evntsel & 0x00010000) {
	fputs(",user", fp);
    }
    if (evntsel & 0x00020000) {
	fputs(",kernel", fp);
    }

    return 0;
}

void do_print(FILE *resfile,
              const struct perfctr_cpu_control *cpu_control,
              const struct perfctr_sum_ctrs *sum)
{
    unsigned int nrctrs, i;
    int (*decoder)(FILE *, unsigned, unsigned) = NULL;

    if (decode_level) {
      decoder = symbolic_decode_event;
    }

    if( cpu_control->tsc_on )
        fprintf(resfile, "tsc\t\t\t%19lld\n", sum->tsc);
    nrctrs = cpu_control->nractrs;
    for (i = 0; i < nrctrs; ++i) {
	if (decoder && decoder(resfile,
			       cpu_control->evntsel[i],
			       ESCR(cpu_control, i)) == 0)
	    goto decoded;
	generic_decode_event(resfile,
			     cpu_control->evntsel[i], ESCR(cpu_control, i));
    decoded:
	fprintf(resfile, "\t%19lld\n", sum->pmc[i]);
    }
#ifndef __x86_64__
    if( cpu_control->p4.pebs_enable )
	fprintf(resfile, "PEBS_ENABLE 0x%08X\n",
		cpu_control->p4.pebs_enable);
    if( cpu_control->p4.pebs_matrix_vert )
	fprintf(resfile, "PEBS_MATRIX_VERT 0x%08X\n",
		cpu_control->p4.pebs_matrix_vert);
#endif
}

void do_arch_usage(void)
{
    if (!am_i_k7()) {
        fprintf(stderr, "\t--p4pe=<value>\t\t\tValue for PEBS_ENABLE (P4 only)\n");
        fprintf(stderr, "\t--p4_pebs_enable=<value>\tSame as --p4pe=<value>\n");
        fprintf(stderr, "\t--p4pmv=<value>\t\t\tValue for PEBS_MATRIX_VERT (P4 only)\n");
        fprintf(stderr, "\t--p4_pebs_matrix_vert=<value>\tSame as --p4pmv=<value>\n");
        fprintf(stderr, "\n");
    }
    fprintf(stderr, "Syntax of event specifiers:\n");
    fprintf(stderr, "\tevent ::= evntsel[/escr][@pmc]\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\tevntsel, escr, and pmc are decimal or hexadecimal numbers.\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\tevntsel is the primary processor-specific event selection code\n");
    fprintf(stderr, "\tto use for this counter. This field is mandatory.\n");
    if (!am_i_k7()) {
        fprintf(stderr, "\tOn a P4, evntsel is written to the counter's CCCR register.\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "\tescr describes the additional event selection data written to\n");
        fprintf(stderr, "\tthe counter's associated ESCR register. (P4 only)\n");
        fprintf(stderr, "\n");
    }
    fprintf(stderr, "\tpmc describes which CPU counter to use for this event.\n");
    fprintf(stderr, "\tBy default the events use counters 0 and up in the order listed.\n");
    if (!am_i_k7()) {
        fprintf(stderr, "\tOn P4, each event is compatible with only a small subset of the\n");
        fprintf(stderr, "\tcounters, and explicit counter assignment is mandatory. Also,\n");
        fprintf(stderr, "\ton P4 bit 31 should be set in pmc to enable 'fast rdpmc'.\n");
    }
    fprintf(stderr, "\tVIA C3 accepts a single event only, but it must use counter 1.\n");
}

extern int parse_event_spec(const char *arg, unsigned int *evntsel,
                            unsigned int *escr, unsigned int *pmc);

unsigned int do_event_spec(unsigned int n,
                           const char *arg,
                           struct perfctr_cpu_control *cpu_control)
{
    unsigned int spec_evntsel, spec_escr, spec_pmc;

    if( parse_event_spec(arg, &spec_evntsel, &spec_escr, &spec_pmc) ) {
        fprintf(stderr, "perfex: invalid event specifier: '%s'\n", arg);
        exit(1);
    }
    if( n >= ARRAY_SIZE(cpu_control->evntsel) ) {
        fprintf(stderr, "perfex: too many event specifiers\n");
        exit(1);
    }
    if( spec_pmc == (unsigned int)-1 )
        spec_pmc = n;
    cpu_control->evntsel[n] = spec_evntsel;
    if (!am_i_k7()) {
        ESCR(cpu_control, n) = spec_escr;
    }
    cpu_control->pmc_map[n] = spec_pmc;
    cpu_control->nractrs = ++n;
    return n;
}

static int parse_value(const char *arg, unsigned int *value)
{
    char *endp;

    *value = strtoul(arg, &endp, 16);
    return endp[0] != '\0';
}

int do_arch_option(int ch,
                   const char *arg,
                   struct perfctr_cpu_control *cpu_control)
{
#ifndef __x86_64__
    unsigned int spec_value;

    switch( ch ) {
    case 1:
	if( parse_value(arg, &spec_value) ) {
	    fprintf(stderr, "perfex: invalid value: '%s'\n", arg);
	    exit(1);
	}
	cpu_control->p4.pebs_enable = spec_value;
	return 0;
    case 2:
	if( parse_value(arg, &spec_value) ) {
	    fprintf(stderr, "perfex: invalid value: '%s'\n", arg);
	    exit(1);
	}
	cpu_control->p4.pebs_matrix_vert = spec_value;
	return 0;
    }
#endif
    return -1;
}
