/* $Id$
 * x86-specific flag parsing code.
 *
 * Copyright (C) 2004 Bryan O'Sullivan
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "libperfctr.h"
#include "arch.h"

struct mask_flag {
    char *name;
    unsigned mask;
};

struct cpu_mask_data {
    unsigned always_or;
    void (*dwim)(unsigned *evntsel);
    int (*parse)(const char *arg, int len, unsigned *evntsel);
    struct mask_flag flags[];
};

static void do_what_i_mean(unsigned *evntsel)
{
    if (!(*evntsel & 0x00030000)) {
        *evntsel |= 0x00010000;
    }
}

static int std_parse(const char *arg, int len, unsigned *evntsel)
{
    if (strncasecmp("unit=", arg, 5) == 0) {
        char *endp;
        unsigned unit;
        unit = strtoul(arg + 5, &endp, 0);
        if (endp == arg + 5 || errno == EINVAL || errno == ERANGE ||
            unit > 255) {
            fprintf(stderr, "invalid unit mask %.*s\n", len - 5, arg + 5);
            exit(1);
        }
        *evntsel |= unit << 8;
        return 0;
    }
    return -1;
}

static const struct cpu_mask_data k7_mask_data = {
    .always_or = 0x00400000,
    .dwim = do_what_i_mean,
    .parse = std_parse,
    .flags = {
        { "user", 0x00010000 },
        { "kernel", 0x00020000 },
        { "invert", 0x00800000 },
        { NULL, 0 },
    },
};

static const struct cpu_mask_data p6_mask_data = {
    .always_or = 0x00400000,
    .dwim = do_what_i_mean,
    .parse = std_parse,
    .flags = {
        { "user", 0x00010000 },
        { "kernel", 0x00020000 },
        { NULL, 0 },
    },
};

struct parse_stuff {
    const char *arg;
    int len;
    unsigned *evntsel;
};

static int parse_event(const struct perfctr_event_set *set,
		       const struct perfctr_event *event,
		       void *thunk)
{
    struct parse_stuff *stuff = thunk;
    
    if (strlen(event->name) == stuff->len &&
        strncasecmp(stuff->arg, event->name, stuff->len) == 0) {
        *stuff->evntsel |= event->evntsel;
        return -1;
    }
    return 0;
}

static int parse_flag(const char *arg, int len, unsigned *evntsel,
                      const struct mask_flag *flags)
{
    const struct mask_flag *flag;
    char *endp;

    for (flag = flags; flag->name; flag++) {
        if (strlen(flag->name) == len &&
            strncasecmp(arg, flag->name, len) == 0) {
            *evntsel |= flag->mask;
            return 0;
        }
    }
    *evntsel |= strtoul(arg, &endp, 0);
    if (endp == arg || errno == EINVAL || errno == ERANGE)
        return -1;
    return 0;
}

static int parse_evntsel(const char *arg, unsigned *evntsel, char **endp)
{
    const struct perfctr_event_set *set;
    const struct cpu_mask_data *data;

    if ((set = perfctr_cpu_event_set(info.cpu_type)) == NULL) {
        perror("perfctr_cpu_event_set");
        goto bail;
    }

    switch (info.cpu_type) {
    case PERFCTR_X86_AMD_K7:
    case PERFCTR_X86_AMD_K8:
    case PERFCTR_X86_AMD_K8C:
        data = &k7_mask_data;
        break;
    case PERFCTR_X86_INTEL_P6:
    case PERFCTR_X86_INTEL_PII:
    case PERFCTR_X86_INTEL_PIII:
    case PERFCTR_X86_INTEL_PENTM:
	data = &p6_mask_data;
	break;
    default:
        data = NULL;
    }

    do {
        int len = strcspn(arg, ",@/");
	struct parse_stuff stuff = { arg, len, evntsel };

        if (perfctr_walk_event_set(set, parse_event, &stuff) == -1) {
            goto parsed;
        }

        if (data) {
            if (parse_flag(arg, len, evntsel, data->flags) == 0) {
                goto parsed;
            }

            if (data->parse && data->parse(arg, len, evntsel) == 0) {
                goto parsed;
            }
        }

        *evntsel |= strtoul(arg, endp, 0);

        if (*endp == arg || errno == EINVAL || errno == ERANGE) {
            goto bail;
        }
    parsed:
        arg += len;
        if (*arg == ',') {
            arg++;
        }
    } while (*arg && *arg != '@' && *arg != '/');

    if (data) {
        *evntsel |= data->always_or;
        if (data->dwim)
            data->dwim(evntsel);
    }

    *endp = (char *) arg;

    return 0;
 bail:
    return -1;
}

int parse_event_spec(const char *arg, unsigned *evntsel,
                            unsigned *escr, unsigned *pmc)
{
    char *endp;

    *evntsel = 0;

    if (parse_evntsel(arg, evntsel, &endp) == -1) {
        return -1;
    }

    if( endp[0] != '/' ) {
        *escr = 0;
    } else {
        arg = endp + 1;
        *escr = strtoul(arg, &endp, 0);
    }
    if( endp[0] != '@' ) {
        *pmc = (unsigned)-1;
    } else {
        arg = endp + 1;
        *pmc = strtoul(arg, &endp, 0);
    }
    return endp[0] != '\0';
}
